#include <stdlib.h>
#include <string.h>

#include "Global.h"
#include "data.h"

#define SUP_MAP_INIT_SIZE   5000
int* gpgraph;
int** gppadj_lists;
int* gptid_sup_map;
int gnmax_vertex_id;
int gnmax_trans_len;
int gndb_size;
int gnmax_sup;

ITEM_NODE *gppattern;
int* gpclosure_buf;
int gnpat_len;

int CountFreqVertices(char* szdataset_name, int nmin_sup, int* &pitem_sup_map);
void LoadGraph(char* szdataset_name, int nmin_sup, int nmin_size, int* pitem_sup_map);
int CountFreqItems(char* szdataset_name, int nmin_size, int* &pitem_sup_map);
void LoadDB(char* szdataset_name, int nmin_sup, int nmin_size, int* pitem_sup_map);
void DestroyDB();

bool IsDuplicate(int* ppattern, int* ptidlist, int length);

void GenTidList(char* szpat_filename, int ntype, int nmin_sup, int nmin_size, char* szout_filename);
int GenOneList(int* ppattern, int npat_len, int npat_sup, int* ptidlist);


void GenTList(char* szdata_filename, int ntype, int nmin_sup, int nmin_size, char* szpat_filename, char* szoutput_filename)
{
	int *pitem_sup_map, i;

	gpgraph = NULL;
	gppadj_lists = NULL;

	pitem_sup_map = NULL;
	if(ntype==0)
	{
		CountFreqItems(szdata_filename, nmin_size, pitem_sup_map);
		LoadDB(szdata_filename, nmin_sup, nmin_size, pitem_sup_map);	
	}
	else
	{
		if(nmin_size>nmin_sup)
		{
			i = nmin_size;
			nmin_size = nmin_sup;
			nmin_sup = i;
		}
		CountFreqVertices(szdata_filename, nmin_sup, pitem_sup_map);
		LoadGraph(szdata_filename, nmin_sup, nmin_size, pitem_sup_map);
	}
	delete []pitem_sup_map;

	gptid_sup_map = new int[gndb_size];
	memset(gptid_sup_map, 0, sizeof(int)*gndb_size);


	GenTidList(szpat_filename, ntype, nmin_sup, nmin_size, szoutput_filename);

	DestroyDB();
	delete []gptid_sup_map;

}

int CountFreqVertices(char* szdataset_name, int nmin_sup, int* &pitem_sup_map)
{
	Data *pdata;
	Transaction* ptransaction;
	int *ptemp_map, ncapacity, nnewcapacity;
	int ntotal_edges;

	pdata = new Data(szdataset_name);

	ncapacity = SUP_MAP_INIT_SIZE;
	pitem_sup_map = new int[ncapacity];
	memset(pitem_sup_map, 0, sizeof(int)*ncapacity);

	gnmax_vertex_id = 0;
	gnmax_trans_len = 0;
	ntotal_edges = 0;

	ptransaction = pdata->getNextTransaction();
	while(ptransaction)
	{
		if(gnmax_vertex_id>=ncapacity)
		{
			nnewcapacity = 2*ncapacity;
			ptemp_map = new int[nnewcapacity];
			memcpy(ptemp_map, pitem_sup_map, sizeof(int)*ncapacity);
			memset(&(ptemp_map[ncapacity]), 0, sizeof(int)*(nnewcapacity-ncapacity));
			delete []pitem_sup_map;
			pitem_sup_map = ptemp_map;
			ncapacity = nnewcapacity;
		}
		pitem_sup_map[gnmax_vertex_id] = ptransaction->length;

		ntotal_edges += ptransaction->length;
		gnmax_vertex_id++;
		if(gnmax_trans_len<ptransaction->length)
			gnmax_trans_len = ptransaction->length;
		ptransaction = pdata->getNextTransaction();
	}
	delete pdata;

	gndb_size = gnmax_vertex_id;

//	printf("%d\n", ntotal_edges);

	return ncapacity;
}

void LoadGraph(char* szdataset_name, int nmin_sup, int nmin_size, int* pitem_sup_map)
{
	Data *pdata;
	Transaction *ptransaction;
	int i, ndb_size, num_of_vertices, nlen, num_of_freqitems;

	gnmax_sup = 0;
	ndb_size = 0;
	num_of_freqitems = 0;
	for(i=0;i<gnmax_vertex_id;i++)
	{
		if(pitem_sup_map[i]>=nmin_sup)
		{
			ndb_size += pitem_sup_map[i];
			num_of_freqitems++;
			if(gnmax_sup < pitem_sup_map[i])
				gnmax_sup = pitem_sup_map[i];
		}
	}
	if(num_of_freqitems<nmin_size)
		return;

	gpgraph = new int[ndb_size+gnmax_vertex_id];
	gppadj_lists = new int*[gnmax_vertex_id];

	num_of_vertices = 0;
	ndb_size = 0;

	pdata = new Data(szdataset_name);
	ptransaction = pdata->getNextTransaction();
	while(ptransaction)
	{
		if(ptransaction->length>=nmin_sup)
		{
			gppadj_lists[num_of_vertices] = &(gpgraph[ndb_size]);
			nlen = 1;
			for(i=0;i<ptransaction->length;i++)
			{
				if(pitem_sup_map[ptransaction->t[i]]>=nmin_size)
					gppadj_lists[num_of_vertices][nlen++] = ptransaction->t[i];
			}
			if(nlen-1>=nmin_sup)
			{
				gppadj_lists[num_of_vertices][0] = nlen-1;
				ndb_size += nlen;
			}
			else
				gppadj_lists[num_of_vertices] = NULL;
		}
		else
			gppadj_lists[num_of_vertices] = NULL;

		num_of_vertices++;
		ptransaction = pdata->getNextTransaction();
	}
	delete pdata;
}


void DestroyDB()
{
	if(gpgraph!=NULL)
		delete []gpgraph;
	if(gppadj_lists!=NULL)
		delete []gppadj_lists;
}


int CountFreqItems(char* szdataset_name, int nmin_size, int* &pitem_sup_map)
{
	Data *pdata;
	Transaction* ptransaction;
	int *ptemp_map, ncapacity, nnewcapacity;
	int i;

	pdata = new Data(szdataset_name);

	ncapacity = SUP_MAP_INIT_SIZE;
	pitem_sup_map = new int[ncapacity];
	memset(pitem_sup_map, 0, sizeof(int)*ncapacity);

	gnmax_vertex_id = 0;
	gnmax_trans_len = 0;
	gndb_size = 0;

	ptransaction = pdata->getNextTransaction();
	while(ptransaction)
	{
		if(ptransaction->length>=nmin_size)
		{
			for(i=0;i<ptransaction->length;i++)
			{
				if(gnmax_vertex_id<ptransaction->t[i])
					gnmax_vertex_id = ptransaction->t[i];

				if(ptransaction->t[i]>=ncapacity)
				{
					nnewcapacity = 2*ncapacity;
					if(nnewcapacity<ptransaction->t[i]+1)
						nnewcapacity = ptransaction->t[i]+1;
					ptemp_map = new int[nnewcapacity];
					memcpy(ptemp_map, pitem_sup_map, sizeof(int)*ncapacity);
					memset(&(ptemp_map[ncapacity]), 0, sizeof(int)*(nnewcapacity-ncapacity));
					delete []pitem_sup_map;
					pitem_sup_map = ptemp_map;
					ncapacity = nnewcapacity;
				}
				pitem_sup_map[ptransaction->t[i]]++;
			}
		}
		if(gnmax_trans_len<ptransaction->length)
			gnmax_trans_len = ptransaction->length;
		gndb_size++;
		ptransaction = pdata->getNextTransaction();
	}
	delete pdata;

	gnmax_vertex_id++;

	return ncapacity;
}

void LoadDB(char* szdataset_name, int nmin_sup, int nmin_size, int* pitem_sup_map)
{
	Data *pdata;
	Transaction *ptransaction;
	int i, ndb_size, num_of_vertices, nvertex, num_of_freqitems;
	int *ppos;

	gnmax_sup = 0;
	ndb_size = 0;
	num_of_freqitems = 0;
	for(i=0;i<gnmax_vertex_id;i++)
	{
		if(pitem_sup_map[i]>=nmin_sup)
		{
			ndb_size += pitem_sup_map[i];
			num_of_freqitems++;
			if(gnmax_sup<pitem_sup_map[i])
				gnmax_sup = pitem_sup_map[i];
		}
		else
			pitem_sup_map[i] = 0;
	}
	if(num_of_freqitems<nmin_size)
		return;

	gpgraph = new int[ndb_size+gnmax_vertex_id];
	gppadj_lists = new int*[gnmax_vertex_id];
	ppos = new int[gnmax_vertex_id];

	ndb_size = 0;
	for(i=0;i<gnmax_vertex_id;i++)
	{
		if(pitem_sup_map[i])
		{
			gppadj_lists[i] = &(gpgraph[ndb_size]);
			ppos[i] = 1;
			ndb_size += pitem_sup_map[i]+1;
		}
		else
			gppadj_lists[i] = NULL;
	}

	num_of_vertices = 0;
	pdata = new Data(szdataset_name);

	ptransaction = pdata->getNextTransaction();
	while(ptransaction)
	{
		if(ptransaction->length>=nmin_size)
		{
			for(i=0;i<ptransaction->length;i++)
			{
				nvertex = ptransaction->t[i];
				if(pitem_sup_map[nvertex])
					gppadj_lists[nvertex][ppos[nvertex]++] = num_of_vertices;
			}
		}
		num_of_vertices++;
		ptransaction = pdata->getNextTransaction();
	}
	delete pdata;

	for(i=0;i<gnmax_vertex_id;i++)
	{
		if(pitem_sup_map[i])
			gppadj_lists[i][0] = ppos[i]-1;
	}

	delete []ppos;

}


void GenTidList(char* szpat_filename, int ntype, int nmin_sup, int nmin_size, char* szout_filename)
{
	FILE *fppat, *fpout;
	int num_of_pats, num_of_bps, npat_len, nsup, *ppattern, i;
	int nlist_len, *ptidlist;

	fppat = fopen(szpat_filename, "rt");
	if(fppat==NULL)
	{
		printf("Error: cannot open file %s for read\n", szpat_filename);
		return;
	}
	fpout = fopen(szout_filename, "wt");
	if(fpout==NULL)
	{
		printf("Error: cannot open file %s for write\n", szout_filename);
		return;
	}

	gppattern = new ITEM_NODE[gnmax_trans_len];
	gpclosure_buf = new int[gnmax_trans_len*gnmax_sup];
	gnpat_len = 0;
	for(i=0;i<gnmax_trans_len;i++)
		gppattern[i].pclosure = &(gpclosure_buf[i*gnmax_sup]);

	ppattern = new int[gnmax_trans_len];
	ptidlist = new int[gnmax_sup];

	num_of_pats = 0;
	num_of_bps = 0;
	fscanf(fppat, "%d", &npat_len);
	while(!feof(fppat))
	{
		for(i=0;i<npat_len;i++)
			fscanf(fppat, "%d", &(ppattern[i]));
		fscanf(fppat, "%d", &nsup);

		if(ntype==0 || nsup>=npat_len)
		{
			if(npat_len>=nmin_size && nsup>=nmin_sup)
			{
				nlist_len = GenOneList(ppattern, npat_len, nsup, ptidlist);

				if(ntype==0 || npat_len!=nlist_len || !IsDuplicate(ppattern, ptidlist, npat_len))
				{
					fprintf(fpout, "%d ", npat_len);
					for(i=0;i<npat_len;i++)
						fprintf(fpout, "%d ", ppattern[i]);
					//fprintf(fpout, "; ");
					fprintf(fpout, "\n");
					fprintf(fpout, "%d ", nlist_len);
					for(i=0;i<nlist_len;i++)
						fprintf(fpout, "%d ", ptidlist[i]);
					fprintf(fpout, "\n");
					num_of_bps++;
				}
			}
		}

		num_of_pats++;
		fscanf(fppat, "%d", &npat_len);
	}

	delete []gpclosure_buf;
	delete []gppattern;

	delete []ppattern;
	delete []ptidlist;

	fclose(fppat);
	fclose(fpout);

	printf("#Maximal bicliques: %d\n", num_of_bps);
	gnum_of_bps = num_of_bps;
}


int GenOneList(int* ppattern, int npat_len, int npat_sup, int* ptidlist)
{
	int nlist_len, nitem, i, j;

	i = 0;
	while(i<gnpat_len && i<npat_len)
	{
		if(ppattern[i]==gppattern[i].nitem)
			i++;
		else
			break;
	}
	if(i>0)
	{
		for(j=0;j<gppattern[i-1].length;j++)
			gptid_sup_map[gppattern[i-1].pclosure[j]] = i;
		gnpat_len = i;
	}
	else 
	{
		nitem = ppattern[0];
		for(j=1;j<=gppadj_lists[nitem][0];j++)
			gptid_sup_map[gppadj_lists[nitem][j]]++;
		gppattern[0].nitem = nitem;
		gppattern[0].length = gppadj_lists[nitem][0];
		memcpy(gppattern[0].pclosure, &(gppadj_lists[nitem][1]), sizeof(int)*gppadj_lists[nitem][0]);
		gnpat_len = 1;
		i++;
	}

	for(j=i;j<npat_len;j++)
	{
		nitem = ppattern[j];
		gppattern[gnpat_len].nitem = nitem;
		gppattern[gnpat_len].length = 0;
		for(i=1;i<=gppadj_lists[nitem][0];i++)
		{
			if(gptid_sup_map[gppadj_lists[nitem][i]]==j)
			{
				gptid_sup_map[gppadj_lists[nitem][i]]++;
				gppattern[gnpat_len].pclosure[gppattern[gnpat_len].length++] = gppadj_lists[nitem][i];
			}
		}
		gnpat_len++;
	}

	nlist_len = 0;
	nitem = ppattern[0];
	for(i=1;i<=gppadj_lists[nitem][0];i++)
	{
		if(gptid_sup_map[gppadj_lists[nitem][i]]==npat_len)
			ptidlist[nlist_len++] = gppadj_lists[nitem][i];
		gptid_sup_map[gppadj_lists[nitem][i]] = 0;
	}

	if(nlist_len!=npat_sup)
		printf("Error: inconsistent support: %d %d\n", nlist_len, npat_sup);

	return nlist_len;
}


bool IsDuplicate(int* ppattern, int* ptidlist, int length)
{
	int i;

	for(i=0;i<length;i++)
	{
		if(ppattern[i]<ptidlist[0])
			return false;
		else if(ppattern[i]==ptidlist[0])
			printf("Error: self-loop?\n");
	}

	return true;
}

